@@ -5,7 +5,13 @@ from pythoncom import com_error
 import robot.libraries.Screenshot as screenshot
 import os
 from robot.api import logger
-from subprocess import check_output,Popen
+import getpass
+from subprocess import check_output, Popen, PIPE
+import win32gui
+import win32gui_struct
+import win32con
+import xlwings as xw
+
 
 class AISAPLibrary:
     """The SapGuiLibrary is a library that enables users to create tests for the Sap Gui application
@@ -49,6 +55,8 @@ class AISAPLibrary:
         self.sapapp = -1
         self.session = -1
         self.connection = -1
+        self.connection_name = None
+        self.window_title = []
 
         self.take_screenshots = screenshots_on_error
         self.screenshot = screenshot.Screenshot()
@@ -106,7 +114,226 @@ class AISAPLibrary:
             self.take_screenshot()
             message = "No existing connection for '%s' found." % connection_name
             raise ValueError(message)
+    
+    def change_session(self, session_num='1'):
+        '''Function to change session'''
+        self.session = self.connection.children(int(session_num))
+     
+    def check_busy_other_session(self,session_num='1'):
+        '''Function to check if session is busy or not'''
+        while True :
+            if self.connection.children(int(session_num)).busy == False :
+                break
+        
+    def get_value_other_session(self, element_id, session_num='1'):
+        element_type = self.get_element_type_other_session(element_id,session_num=session_num)
+        return_value = ""
+        if (element_type == "GuiTextField"
+                or element_type == "GuiCTextField"
+                or element_type == "GuiLabel"
+                or element_type == "GuiTitlebar"
+                or element_type == "GuiStatusbar"
+                or element_type == "GuiButton"
+                or element_type == "GuiTab"
+                or element_type == "GuiShell"):
+            self.set_focus_other_session(element_id,session_num=session_num)
+            return_value = self.connection.children(int(session_num)).findById(element_id).text
+        elif element_type == "GuiStatusPane":
+            return_value = self.connection.children(int(session_num)).findById(element_id).text
+        elif (element_type == "GuiCheckBox"
+              or element_type == "GuiRadioButton"):
+            actual_value = self.connection.children(int(session_num)).findById(element_id).selected
+            # In these situations we return check / unchecked, so we change these values here
+            if actual_value == True:
+                return_value = "checked"
+            elif actual_value == False:
+                return_value = "unchecked"
+        elif element_type == "GuiComboBox":
+            return_value = self.connection.children(int(session_num)).findById(element_id).text
+            # In comboboxes there are many spaces after the value. In order to check the value, we strip them away.
+            return_value = return_value.strip()
+        else:
+            # If we can't return the value for this element type, raise an assertion error
+            self.take_screenshot()
+            message = "Cannot get value for element type '%s'" % element_type
+            raise Warning(message)
+        return return_value
+        
+    def element_value_should_contain_other_session(self, element_id, expected_value, message=None, session_num='1'):
+        element_type = self.get_element_type_other_session(element_id,session_num=session_num)
 
+        # Breaking up the different element types so we can check the value the correct way
+        if (element_type == "GuiTextField"
+                or element_type == "GuiCTextField"
+                or element_type == "GuiComboBox"
+                or element_type == "GuiLabel"):
+            self.connection.children(int(session_num)).findById(element_id).setfocus()
+            actual_value = self.get_value_other_session(element_id, session_num=session_num)
+            time.sleep(self.explicit_wait)
+            # In these cases we can simply check the text value against the value of the element
+            if expected_value not in actual_value:
+                self.take_screenshot()
+                if message is None:
+                    message = "Element value '%s' does not contain '%s', (but was '%s')" % (
+                        element_id, expected_value, actual_value)
+                raise AssertionError(message)
+        else:
+            # When the element content can't be checked, raise an assertion error
+            self.take_screenshot()
+            message = "Cannot use keyword 'element value should contain' for element type '%s'" % element_type
+            raise Warning(message)
+        # Run explicit wait as last
+        time.sleep(self.explicit_wait)
+        
+    def set_focus_other_session(self, element_id, session_num='1'):
+
+        """Sets the focus to the given element.
+        """
+        element_type = self.get_element_type_other_session(element_id,session_num=session_num)
+        if element_type != "GuiStatusPane":
+            self.connection.children(int(session_num)).findById(element_id).setFocus()
+        time.sleep(self.explicit_wait)
+        
+    def doubleclick_element_other_session(self, element_id, item_id, column_id, session_num='1'):
+        """Performs a double-click on a given element. Used only for shell objects.
+        """
+
+        # Performing the correct method on an element, depending on the type of element
+        element_type = self.get_element_type_other_session(element_id,session_num=session_num)
+        if element_type == "GuiShell":
+            self.connection.children(int(session_num)).findById(element_id).doubleClickItem(item_id, column_id)
+        else:
+            self.take_screenshot()
+            message = "You cannot use 'doubleclick element' on element type '%s', maybe use 'click element' instead?" % element_type
+            raise Warning(message)
+        time.sleep(self.explicit_wait)
+        
+    def click_element_other_session(self, element_id, session_num='1'):
+        """Performs a single click on a given element. Used only for buttons, tabs and menu items.
+
+        In case you want to change a value of an element like checkboxes of selecting an option in dropdown lists,
+        use `select checkbox` or `select from list by label` instead.
+        """
+
+        # Performing the correct method on an element, depending on the type of element
+        element_type = self.get_element_type_other_session(element_id,session_num=session_num)
+        if (element_type == "GuiTab"
+                or element_type == "GuiMenu"):
+            self.connection.children(int(session_num)).findById(element_id).select()
+        elif element_type == "GuiButton":
+            self.connection.children(int(session_num)).findById(element_id).press()
+        else:
+            self.take_screenshot()
+            message = "You cannot use 'click_element' on element type '%s', maybe use 'select checkbox' instead?" % element_type
+            raise Warning(message)
+        time.sleep(self.explicit_wait)
+        
+    def get_element_type_other_session(self, element_id, session_num='1'):
+        """Returns the Sap element type for the given element.
+        """
+        try:
+            type = self.connection.children(int(session_num)).findById(element_id).type
+            return type
+        except com_error:
+            self.take_screenshot()
+            message = "Cannot find element with id '%s'" % element_id
+            raise ValueError(message)
+        
+    def input_text_other_session(self, element_id, text,  session_num='1'):
+        """Inserts the given text into the text field identified by locator.
+        Use keyword `input password` to insert a password in a text field.
+        """
+        element_type = self.get_element_type_other_session(element_id,session_num=session_num)
+        if (element_type == "GuiTextField"
+                or element_type == "GuiCTextField"
+                or element_type == "GuiShell"
+                or element_type == "GuiPasswordField"):
+            self.connection.children(int(session_num)).findById(element_id).text = text
+            logger.info("Typing text '%s' into text field '%s'." % (text, element_id))
+            time.sleep(self.explicit_wait)
+        else:
+            self.take_screenshot()
+            message = "Cannot use keyword 'input text' for element type '%s'" % element_type
+            raise ValueError(message)
+        
+    def select_radio_button_other_session(self, element_id, session_num='1'):
+        """Sets radio button to the specified value.
+        """
+        element_type = self.get_element_type_other_session(element_id,session_num=session_num)
+        if element_type == "GuiRadioButton":
+            self.connection.children(int(session_num)).findById(element_id).selected = True
+        else:
+            self.take_screenshot()
+            message = "Cannot use keyword 'select radio button' for element type '%s'" % element_type
+            raise ValueError(message)
+        time.sleep(self.explicit_wait)
+        
+    def run_transaction_other_session(self, transaction, session_num='1'):
+        self.connection.children(int(session_num)).findById("wnd[0]/tbar[0]/okcd").text = transaction
+        time.sleep(self.explicit_wait)
+        self.send_vkey_other_session(0, session_num=session_num)
+
+        if transaction == '/nex':
+            return
+
+        pane_value = self.connection.children(int(session_num)).findById("wnd[0]/sbar/pane[0]").text
+        if pane_value in ("Transactie %s bestaat niet" % transaction.upper(),
+                          "Transaction %s does not exist" % transaction.upper(),
+                          "Transaktion %s existiert nicht" % transaction.upper()):
+            self.take_screenshot()
+            message = "Unknown transaction: '%s'" % transaction
+            raise ValueError(message)
+        
+    def send_vkey_other_session(self, vkey_id, window=0, session_num='1'):
+        vkey_id = str(vkey_id)
+        vkeys_array = ["ENTER", "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "F10", "F11", "F12",
+                       None, "SHIFT+F2", "SHIFT+F3", "SHIFT+F4", "SHIFT+F5", "SHIFT+F6", "SHIFT+F7", "SHIFT+F8",
+                       "SHIFT+F9", "CTRL+SHIFT+0", "SHIFT+F11", "SHIFT+F12", "CTRL+F1", "CTRL+F2", "CTRL+F3", "CTRL+F4",
+                       "CTRL+F5", "CTRL+F6", "CTRL+F7", "CTRL+F8", "CTRL+F9", "CTRL+F10", "CTRL+F11", "CTRL+F12",
+                       "CTRL+SHIFT+F1", "CTRL+SHIFT+F2", "CTRL+SHIFT+F3", "CTRL+SHIFT+F4", "CTRL+SHIFT+F5",
+                       "CTRL+SHIFT+F6", "CTRL+SHIFT+F7", "CTRL+SHIFT+F8", "CTRL+SHIFT+F9", "CTRL+SHIFT+F10",
+                       "CTRL+SHIFT+F11", "CTRL+SHIFT+F12", None, None, None, None, None, None, None, None, None, None,
+                       None, None, None, None, None, None, None, None, None, None, None, "CTRL+E", "CTRL+F", "CTRL+A",
+                       "CTRL+D", "CTRL+N", "CTRL+O", "SHIFT+DEL", "CTRL+INS", "SHIFT+INS", "ALT+BACKSPACE",
+                       "CTRL+PAGEUP", "PAGEUP",
+                       "PAGEDOWN", "CTRL+PAGEDOWN", "CTRL+G", "CTRL+R", "CTRL+P", "CTRL+B", "CTRL+K", "CTRL+T",
+                       "CTRL+Y",
+                       "CTRL+X", "CTRL+C", "CTRL+V", "SHIFT+F10", None, None, "CTRL+#"]
+
+        # If a key combi is given, replace vkey_id by correct id based on given combination
+        if not vkey_id.isdigit():
+            search_comb = vkey_id.upper()
+            search_comb = search_comb.replace(" ", "")
+            search_comb = search_comb.replace("CONTROL", "CTRL")
+            search_comb = search_comb.replace("DELETE", "DEL")
+            search_comb = search_comb.replace("INSERT", "INS")
+            try:
+                vkey_id = vkeys_array.index(search_comb)
+            except ValueError:
+                if search_comb == "CTRL+S":
+                    vkey_id = 11
+                elif search_comb == "ESC":
+                    vkey_id = 12
+                else:
+                    message = "Cannot find given Vkey, provide a valid Vkey number or combination"
+                    raise ValueError(message)
+
+        try:
+            self.connection.children(int(session_num)).findById("wnd[% s]" % window).sendVKey(vkey_id)
+        except com_error:
+            self.take_screenshot()
+            message = "Cannot send Vkey to given window, is window wnd[% s] actually open?" % window
+            raise ValueError(message)
+        time.sleep(self.explicit_wait)
+        
+    def click_element_other_session_process(self, element_id, session_num='1'):
+        ''' click element to other session in multithread or different process '''
+        Popen(["python", os.path.dirname(os.path.realpath(__file__))+'\\click_element_multithread.py',self.connection_name,element_id,session_num])
+        
+    def send_vkey_other_session_process(self, vkey_id, window=0, session_num='1'):
+        ''' send vkey to other session in multithread or different process '''
+        Popen(["python", os.path.dirname(os.path.realpath(__file__))+'\\send_vkey_multithread.py',self.connection_name,vkey_id,str(window),session_num])
+        
     def connect_to_session(self, explicit_wait=0):
         """Connects to an open session SAP.
 
@@ -771,6 +998,12 @@ class AISAPLibrary:
             message = "Cannot use keyword 'unselect checkbox' for element type '%s'" % element_type
             raise ValueError(message)
         time.sleep(self.explicit_wait)
+        
+    def get_all_window_title(self):
+        '''Fungsi untuk mendapatkan title dari window semua session'''
+        children_count = self.connection.children.count
+        for child_num in range(children_count):
+            self.window_title.append(win32gui.GetWindowText(self.connection.children(child_num).FindById("wnd[0]").Handle))
     
     def close_sap(self):
         '''Fungsi untuk melakukan close SAP
@@ -779,21 +1012,36 @@ class AISAPLibrary:
                 Close Sap
         '''
         
-        for i in range(3):
+        children_count = self.connection.children.count
+        for child_num in range(children_count):
+            
+            if len(self.window_title)>0:
+                if self.connection.children(children_count-child_num-1).busy :   
+                    window_title = self.window_title[children_count-child_num-1]
+                    hmenu = win32gui.GetSystemMenu(win32gui.FindWindow("SAP_FRONTEND_SESSION",window_title), False)
+                    buf, extras = win32gui_struct.EmptyMENUITEMINFO(win32con.MIIM_ID )
+                    win32gui.GetMenuItemInfo(hmenu,9,True,buf)
+                    data = win32gui_struct.UnpackMENUITEMINFO(buf)
+
+                    handle = win32gui.FindWindow("SAP_FRONTEND_SESSION",window_title)
+                    win32gui.SendMessage(handle, win32con.WM_ACTIVATE, win32con.WA_ACTIVE, 0)
+                    win32gui.SendMessage(handle, win32con.WM_SYSCOMMAND, data.wID, 0)         
+                
+            for i in range(3):
+                try:
+                    self.send_vkey_other_session(12,0,str(children_count-child_num-1))
+                except:
+                    pass
+            
             try:
-                self.send_vkey(12)
+                self.run_transaction_other_session('/nexit',str(children_count-child_num-1))
+            except:
+                pass
+            
+            try:
+                self.send_vkey_other_session(15,0,str(children_count-child_num-1))
             except:
                 pass
-        
-        try:
-            self.run_transaction('/nexit')
-        except:
-            pass
-        
-        try:
-            self.send_vkey(15)
-        except:
-            pass
         
         try:
             self.click_element('wnd[1]/usr/btnSPOP-OPTION1')
@@ -813,6 +1061,8 @@ class AISAPLibrary:
                 self.session.findById(shell_element).expandNode("         "+node_num)                
             if interactions=='doubleclick':
                 self.session.findById(shell_element).doubleClickNode("         "+node_num)
+            if interactions=='presstoolbar':
+                self.session.findById(shell_element).pressToolbarButton(str(node_num))
         except Exception as e:
             raise RuntimeError(e)
 
@@ -831,7 +1081,49 @@ class AISAPLibrary:
         except Exception as e:
             raise RuntimeError(e)
         
-    def login_sap(self, config):
+    def export_data(self, options=None, format=None, folder_loc=None, file_name=None):
+        if options==None :
+            raise RuntimeError('Please select option, such as excel with format xlsx or xxl')
+        elif options=='excel':
+            self.send_vkey(43)
+            
+            if format=="xxl":
+                self.session.findById("wnd[1]/usr/cmbG_LISTBOX").key = "33"
+                self.send_vkey(0)
+                
+                try :
+                    self.session.findById('wnd[1]/usr/txtMESSTXT2')
+                    self.send_vkey(0)
+                except :
+                    pass
+                
+                self.select_radio_button('wnd[1]/usr/subSUBSCREEN_STEPLOOP:SAPLSPO5:0150/sub:SAPLSPO5:0150/radSPOPLI-SELFLAG[0,0]')
+                    
+                self.send_vkey(0)
+                self.send_vkey(0)
+                wb = xw.Book('Worksheet in Basis (1)',update_links=False)
+                wb.save(folder_loc+file_name)
+                wb.close()
+                Popen("Taskkill /im excel.exe"+' /FI "USERNAME eq '+getpass.getuser(), stdout=PIPE)
+                self.send_vkey(0)
+                
+            if format=='xlsx':
+                self.session.findById("wnd[1]/usr/cmbG_LISTBOX").key = "31"
+                self.send_vkey(0)
+                
+                self.input_text('wnd[1]/usr/ctxtDY_FILENAME',file_name)
+                self.input_text('wnd[1]/usr/ctxtDY_PATH',folder_loc)
+                self.click_element('wnd[1]/tbar[0]/btn[0]')
+                self.send_vkey(12)
+                try :
+                    time.sleep(5)
+                    wb = xw.Book(file_name,update_links=False)
+                    wb.close()
+                    Popen("Taskkill /im excel.exe"+' /FI "USERNAME eq '+getpass.getuser(), stdout=PIPE)
+                except :
+                    pass
+        
+    def login_sap(self, config, time_wait=0):
         """ Script untuk Login SAP
             mengembalikan variabel status (boolean) dan message (string)
             
@@ -865,14 +1157,16 @@ class AISAPLibrary:
             raise RuntimeError('Cannot connect to session')
 
         self.open_connection(config['connection'])
+        
+        self.connection_name = config['connection']
+        
         self.input_text("wnd[0]/usr/txtRSYST-BNAME",config["username"])
         self.input_password("wnd[0]/usr/pwdRSYST-BCODE",config["password"])
         self.input_text("wnd[0]/usr/txtRSYST-MANDT",config['client'])
         self.send_vkey(0)
 
-        time.sleep(5)
-        
-        
+        time.sleep(time_wait)
+          
         msg = "Success Login"
         
         try:
